#module  MAP_UNIBUS_SPACE  "V2.0"

/*|++MAPIOSPAC.C

 Facility: 
	Fermilab Accelerator Control System - VAX/VMS ACNET
	General library routine for use in suitably privileged programs.

 Abstract:
	Maps the UNIBUS space on the specified UBA (UNIBUS Adapter) into
	the virtual address space of the current process as a private section.
	The process must have the PFNMAP privilege and must use great care
	in dealing with the UNIBUS addresses to avoid clobbering an active
	device or attempting to access non-existant addresses.

 Environment:
	User mode from a privileged process or image.
|--*/
/*
 Modification History:
 Author: F. Nagy			Creation date: 08-Sep-83

 V1.0	08-Sep-83  FJN	Created for use in ACLKWATCH (to access UCD) and for
			eventual use in ACNAUX (for rebooting PDP-11's), etc.
 V2.0	15-Jul-86  FJN	Modified to handle 8600/8650 and MicroVAXes
*/

/******************
 * Include Files: *
 ******************/
#include  vaxtypes			/* VAX-11 C Type extensions */
#include  descrip			/* VAX-11 Descriptors */
#include  descrip2			/* Additional descriptor defs. */
#include  secdef			/* Create/map section flags */
#include  syidef			/* Item codes for $GETSYI */
#include  prdef				/* Processor type codes */

/********************************
 * Local definitions and types: *
 ********************************/
typedef unsigned long status;		/* Condition code status type */

					/* Test status for failure and STOP
					   if so, signalling the error */
#define  FAIL_TEST(status)  if ($VMS_FAILURE( status)) LIB$STOP( status)

globalvalue LIB$_INVARG;		/* Invalid argument */
globalvalue LIB$_FATERRLIB;		/* Fatal error in library routine */
globalvalue SS$_BADPARAM;		/* Bad parameter */

/*
 Definitions of the physical addresses of the I/O spaces of the UNIBUS
 Adapters for the various types of VAX's.
*/

					/* VAX-8600,8650 processors: */
#define  IO790$AL_IOA0	 0X20000000	/* Start of I/O space for SBI0 */
#define  IO790$AL_IOA1	 0X22000000	/* Start of I/O space for SBI1 */
#define  IO790$AL_UB0SP  0X20100000	/* UNIBUS 0 I/O space */
#define  IO790$AL_UB1SP  0X20140000	/* UNIBUS 1 I/O space */
#define  IO790$AL_UB2SP  0X20180000	/* UNIBUS 2 I/O space */

					/* VAX-11/780,785 processors: */
#define  IO780$AL_IOBASE 0X20000000	/* Start of I/O space */
#define  IO780$AL_UB0SP  0X20100000	/* UNIBUS 0 I/O space */
#define  IO780$AL_UB1SP  0X20140000	/* UNIBUS 1 I/O space */
#define  IO780$AL_UB2SP  0X20180000	/* UNIBUS 2 I/O space */
#define  IO780$AL_UB3SP  0X201C0000	/* UNIBUS 3 I/O space */

					/* VAX-11/750 processors: */
#define  IO750$AL_IOBASE 0X00F20000	/* Start of I/O space */
#define  IO750$AL_UB0SP  0X00FC0000	/* UNIBUS 0 I/O space */
#define  IO750$AL_UB1SP  0X00F80000	/* UNIBUS 1 I/O space */

					/* VAX-11/730 processors: */
#define  IO730$AL_IOBASE 0X00F20000	/* Start of I/O space */
#define  IO730$AL_UB0SP  0X00FC0000	/* UNIBUS I/O space */

					/* MicroVAX I and II processors: */
#define  IOUV1$AL_QB0SP	 0X20000000	/* MicroVAX I Q22 bus space */
#define  IOUV2$AL_QB0SP	 0X20000000	/* MicroVAX II Q22 bus space */

#define  IO$L_UBALEN	 0X00040000	/* Length of UNIBUS I/O space */
#define  IO$L_Q22LEN	 0X00400000	/* Length of Q22 Bus I/O space */
#define  MAPLEN_MIN	 1024		/* Minimum UNIBUS map length in bytes */

#define  PAGE_SIZE	 512		/* Size of VAX page in bytes */
#define  PAGE_BITS	 9		/* Number of bits in page size */

/**********************************************
 * Global Procedure and Function definitions: *
 **********************************************/
status LIB$GET_EF();			/* Get free event flag */
status LIB$FREE_EF();			/* Release event flag */

status SYS$GETSYI();			/* Get System Information service */
status SYS$CRMPSC();			/* Create and Map Section service */
status SYS$WAITFR();			/* Wait for event flag service */

int LIB_NARGS();			/* Get number of call arguments */


/*+/MAP_IO_SPACE

 Functional Description:
	Map the UNIBUS or QBus I/O and memory space of a specified UNIBUS
	Adapter (or of the Q22 Bus), returning the virtual address of the
	start of the section mapping the UNIBUS/Q22 addresses.

 Input Parameters:
	uba - By value, the UNIBUS/Q22 Adapter to be mapped.  The allowed
	      ranges are (by CPU type):
			86xx's:	0-2 (really 7 but #3 address not fixed!)
			78x's:	0-3
			750's:	0-1
			730's:	0 only
			MicroVAXes: 0 only (Q22 bus)
	      For all other CPU types, this routine returns an error.
	base - By value, the base address in UNIBUS/Q22 space to be mapped.
	       The accepted ranges are (memory plus I/O space):
		    0 to 3FFFF(16) for 256KB UNIBUS
		    0 to 3FFFFF(16) for 4 MB Q22 Bus
	       Default value is 0.  Usually this will be the CSR offset in
	       UNIBUS/Q22 address space.
	length - By value, the number of bytes to be mapped in UNIBUS/Q22 space.
		 The default value is 1024 (2 pages).  The allowed range is
		 1 to 40000(16) for UNIBUS and 1 to 400000(16) for Q22, but
		 the minimum mapped is always at least 2 pages (1024 bytes).

 Implicit Inputs:
	NONE

 Output Parameters:
	Returns the virtual address mapping the specified UNIBUS/Q22 base
	address.

 Implicit Outputs:
	NONE

 Condition Codes:
	All errors are signalled.  Specific signals which may result are:
		LIB$_INSEF	no free event flags for internal use
		LIB$_INVARG	illegal UBA number specified
		LIB$_FATERRLIB	unknown CPU type encountered
		SS$_BADPARAM	illegal UNIBUS base or length specified
		SS$_INSFWSL	insufficient working set limit to map pages
		SS$_NOPRIV	no PFNMAP privilege
		SS$_VASFULL	virtual address space full

 Side Effects:
	Process must have PFNMAP privilege to use this routine.
	Routine comprehends 86xx, 78x, 750, 730 and MicroVAX CPU's at this time.
-*/

address MAP_IO_SPACE( uba, base, length )
int uba;				/* UNIBUS/Q22 Adapter number 0-n */
ulong base;				/* UNIBUS/Q22 Base address 0-3FFFFF */
long length;				/* UNIBUS/Q22 map length 0-400000/-1 */
{
status sts;				/* Completion status code */
long cputype;				/* CPU identification */
struct {				/* Item list for $GETSYI */
    word buflen;			/* Buffer length */
    word itmcod;			/* Item code */
    long *buffer;			/* Pointer to buffer */
    long *rtnlen;			/* Return data length pointer */
    long finis;				/* List terminator */
    } itmlst;
long efn;				/* Event flag number */
uword iosb[4];				/* I/O status block */
ulong pa;				/* UBA physical address */
int ubano = 0;				/* UBA number to map */
ulong off = 0;				/* Offset in UNIBUS addresses */
ulong mpl = MAPLEN_MIN;			/* Number of bytes to map from UNIBUS */
int nn;
address range[2] = { 0, 0 };		/* Section virtual address range */

/*
  Get event flag in case $GETSYI is asynchronous.
*/
sts = LIB$GET_EF( &efn);
FAIL_TEST( sts);

/*
  Initialize item list structure for $GETSYI.
*/
itmlst.buflen = 4;
itmlst.itmcod = SYI$_CPU;
itmlst.buffer = &cputype;
itmlst.rtnlen = 0;
itmlst.finis = 0;

/*
  Get CPU identification even if asynchronous operation.
*/
sts = SYS$GETSYI( efn, 0, 0, &itmlst, iosb, 0, 0);
FAIL_TEST( sts);
sts = SYS$WAITFR( efn);
FAIL_TEST( sts);
FAIL_TEST( iosb[0]);

/*
  Got CPU id, release event flag as not needed anymore.
*/
sts = LIB$FREE_EF( &efn);
FAIL_TEST( sts);

/*
 Set UNIBUS/Q22 length based on CPU type for argument checks.
*/
if ((cputype == PR$_SID_TYPUV1) || (cputype == PR$_SID_TYPUV2))
    pa = IO$L_Q22LEN;
else
    pa = IO$L_UBALEN;

/*
  Now get arguments and check for validity.
*/
switch (nn = LIB_NARGS())
    {
default:				/* >=3 call parameters */
    if (length > pa)
	LIB$STOP( SS$_BADPARAM);	/* Requested length is too long */
    mpl = (length > MAPLEN_MIN) ?
	  length + PAGE_SIZE - 1 :	/* Request rounded to full page */
	  MAPLEN_MIN;			/* Use minimum mapping length */

case 2:					/* Just uba and base parameters */
    off = base;
    if (off >= pa)
	LIB$STOP( SS$_BADPARAM);	/* Request base is out of range */
    if ((nn > 2) && ((off + length) > pa))
	    LIB$STOP( SS$_BADPARAM);	/* Base+length is out of range */
    if (off >= (pa - PAGE_SIZE))
	mpl = PAGE_SIZE;		/* Only final page to be mapped */

case 1:					/* Just uba parameter */
    ubano = uba;

case 0: ;				/* No parameters */
    }

/*
  Now get physical address of UBA address space from CPU type and UBA number.
  Also check for UBA number in legal range for CPU type.
*/
switch (cputype)
    {
case PR$_SID_TYP790:			/* VAX-8600 and 8650 */
    switch (ubano)
	{
    case 0:
	pa = IO790$AL_UB0SP;		/* UB0 on 86xx */
	break;
    case 1:
	pa = IO790$AL_UB1SP;		/* UB1 on 86xx */
	break;
    case 2:
	pa = IO790$AL_UB2SP;		/* UB2 on 86xx */
	break;
    default:
	LIB$STOP( LIB$_INVARG);		/* Only UB0-2 OK here on 86xx */
	}
    break;

case PR$_SID_TYP780:			/* VAX-11/780, 782 and 785 */
    switch (ubano)
	{
    case 0:
	pa = IO780$AL_UB0SP;		/* UB0 on 78x */
	break;
    case 1:
	pa = IO780$AL_UB1SP;		/* UB1 on 78x */
	break;
    case 2:
	pa = IO780$AL_UB2SP;		/* UB2 on 78x */
	break;
    case 3:
	pa = IO780$AL_UB3SP;		/* UB3 on 78x */
	break;
    default:
	LIB$STOP( LIB$_INVARG);		/* Only UB0-3 on 78x */
	}
    break;

case PR$_SID_TYP750:			/* VAX-11/750 */
    if (ubano == 0)
	pa = IO750$AL_UB0SP;		/* 750 has UB0 and */
    else if (ubano == 1)
	pa = IO750$AL_UB1SP;		/* ... UB1 */
    else
	LIB$STOP( LIB$_INVARG);
    break;

case PR$_SID_TYP730:			/* VAX-11/730 */
    if (ubano == 0)
	pa = IO730$AL_UB0SP;		/* 730 only has single UBA */
    else
	LIB$STOP( LIB$_INVARG);
    break;

case PR$_SID_TYPUV1:			/* MicroVAX-I */
case PR$_SID_TYPUV2:			/* MicroVAX-II */
    if (ubano == 0)
	pa = IOUV2$AL_QB0SP;		/* Map Q22 bus I/O space */
    else
	LIB$STOP( LIB$_INVARG);
    break;

default:				/* Unknown CPU */
    LIB$STOP( LIB$_FATERRLIB);
    }

/*
  Now create and map a private section to the Page Frame Numbers given by
  the UNIBUS/Q22 physical addresses in the range selected.
*/
sts = SYS$CRMPSC( &range, &range, 0, SEC$M_EXPREG|SEC$M_WRT|SEC$M_PFNMAP,
		  0, 0, 0, 0, mpl >> PAGE_BITS, (pa + off) >> PAGE_BITS, 0, 0);
FAIL_TEST( sts);

/*
  Return virtual address of the selected UNIBUS/Q22 base address by taking
  the virtual address of the page (range[0]) and adding to it the byte offset
  within that page of the UNIBUS/Q22 base address (off, where PAGE_SIZE-1 is
  the mask for the address bits within a single VAX page).
*/
return range[0] + (off & (PAGE_SIZE - 1));

}
